import os.path
from dataclasses import dataclass, make_dataclass, fields
from omegaconf import OmegaConf as oc
from typing import List, Dict, Any
from .converters import as_time, as_freq
from scabha import schema_utils
from scabha.cargo import Parameter


class BaseConfigSection:
    """Base class for dynamically generated config dataclasses.

    Also implements specific post-init methods for them.
    """

    def validate_choice_fields(self):
        for fld in fields(self):
            name = fld.name
            value = getattr(self, name)
            meta = fld.metadata

            # Optional field values can be None, validated at construction.
            if value is None:
                continue
            elif "choices" in meta:  # Check for choices.
                choices = meta.get("choices")
                assert value in choices, \
                    f"Invalid input in {fld.name}. " \
                    f"User specified '{value}'. " \
                    f"Valid choices are {choices}."
            elif "element_choices" in meta:  # Check for element choices.
                element_choices = meta.get("element_choices")
                if isinstance(value, List):
                    args = value
                elif isinstance(value, Dict):
                    args = value.values()
                else:
                    raise ValueError(f"Paramter {name} of type {type(value)}"
                                     f"has element choices. Not understood.")
                invalid = set(args) - set(element_choices)
                assert not invalid, \
                    f"Invalid input in {fld.name}. " \
                    f"User specified '{','.join(map(str, invalid))}'. " \
                    f"Valid choices are {','.join(map(str,element_choices))}."

    def __input_ms_post_init__(self):

        self.validate_choice_fields()
        self.time_chunk = as_time(self.time_chunk)
        self.freq_chunk = as_freq(self.freq_chunk)

        assert len(self.select_uv_range) == 2, \
            "input_ms.select_uv_range expects a two-element list."

        assert not (self.sigma_column and self.weight_column), \
            "sigma_column and weight_column are mutually exclusive."

        if self.is_bda:
            assert self.time_chunk == 0, \
                ("input_ms.is_bda does not support chunking in time. Please "
                 "set input_ms.time_chunk to 0.")
            assert self.freq_chunk == 0, \
                ("input_ms.is_bda does not support chunking in freq. Please "
                 "set input_ms.freq_chunk to 0.")

    def __input_model_post_init__(self):
        pass

    def __output_post_init__(self):

        self.validate_choice_fields()
        assert not (bool(self.products) ^ bool(self.columns)), \
            "Neither or both of products and columns must be specified."
        if self.products:
            assert len(self.products) == len(self.columns), \
                    "Number of products not equal to number of columns."

        if self.net_gains:
            nested = any(isinstance(i, list) for i in self.net_gains)
            if nested:
                assert all(isinstance(i, list) for i in self.net_gains), \
                    ("Contents of outputs.net_gains not understood. "
                     "Must be strictly a list or list of lists.")
            else:
                assert all(isinstance(i, str) for i in self.net_gains), \
                    ("Contents of outputs.net_gains not understood. "
                     "Must be strictly a list or list of lists.")
                # In the non-nested case, introduce outer list (consistent).
                self.net_gains = [self.net_gains]

    def __mad_flags_post_init__(self):
        pass

    def __solver_post_init__(self):

        self.validate_choice_fields()
        assert len(self.iter_recipe) >= len(self.terms), \
            "User has specified solver.iter_recipe with too few elements."

        assert self.convergence_criteria >= 1e-8, \
            "User has specified solver.convergence_criteria below 1e-8."

    def __dask_post_init__(self):
        self.validate_choice_fields()
        if self.address:
            msg = (
                "Scheduler address supplied but dask.scheduler has not "
                "been set to distributed."
            )
            assert self.scheduler == "distributed", msg

    def __gain_post_init__(self):
        self.validate_choice_fields()
        self.time_interval = as_time(self.time_interval)
        self.freq_interval = as_freq(self.freq_interval)
        if self.type == "crosshand_phase" and self.solve_per != "array":
            raise ValueError("Crosshand phase can only be solved as a per "
                             "array term. Please set the appropriate "
                             "term.solve_per to 'array'.")


@dataclass
class _GainSchema(object):
    gain: Dict[str, Parameter]


# load schema files
_schema = _gain_schema = None

if _schema is None:
    dirname = os.path.dirname(__file__)

    # Make dataclass based on the argument schema.
    _schema = oc.load(f"{dirname}/argument_schema.yaml")

    # Map __post_init__ methods to sections.
    post_init_map = {s: getattr(BaseConfigSection, f"__{s}_post_init__")
                     for s in _schema.keys()}

    # Create the base config class.
    BaseConfig = schema_utils.nested_schema_to_dataclass(
        _schema,
        "BaseConfig",
        section_bases=(BaseConfigSection,),
        post_init_map=post_init_map
    )

    # Create a mapping from section name to section dataclass.
    _config_dataclasses = {f.name: f.type for f in fields(BaseConfig)}

    # The gain section is loaded explicitly, since we need to form up multiple
    # instances.
    _gain_schema = oc.merge(
        oc.structured(_GainSchema),
        oc.load(f"{dirname}/gain_schema.yaml")
    )
    _gain_schema = _gain_schema.gain

    # Create gain dataclass.
    Gain = schema_utils.schema_to_dataclass(
        _gain_schema,
        "Gain",
        bases=(BaseConfigSection,),
        post_init=BaseConfigSection.__gain_post_init__
    )


def get_config_sections():
    """Returns list of known config sections."""
    return list(_config_dataclasses.keys()) + ["gain"]


def finalize_structure(additional_config):

    terms = None

    for cfg in additional_config[::-1]:
        terms = oc.select(cfg, "solver.terms")
        if terms is not None:
            break

    # Use the default terms if no alternative is specified.
    terms = terms or _config_dataclasses["solver"]().terms

    FinalConfig = make_dataclass(
        "FinalConfig",
        [(t, Gain, Gain()) for t in terms],
        bases=(BaseConfig,)
    )

    return FinalConfig


def make_stimela_schema(params: Dict[str, Any],
                        inputs: Dict[str, Parameter],
                        outputs: Dict[str, Parameter]):
    """Augments a schema for stimela based on solver.terms."""

    inputs = inputs.copy()

    if 'solver' in params:
        terms = params['solver'].get('terms', None)
    else:
        terms = params.get('solver.terms', None)
    if terms is None:
        terms = _config_dataclasses["solver"]().terms

    for jones in terms:
        for key, value in _gain_schema.items():
            inputs[f"{jones}.{key}"] = value

    return inputs, outputs



#!/usr/bin/env python

import sys
import click
import os.path
import subprocess
from typing import List, Dict, Optional, Any
from dataclasses import dataclass
from omegaconf import OmegaConf
from rich import print

DEFAULT_MANIFEST = os.path.join(os.path.dirname(__file__), "cargo-manifest.yml")

BASE_IMAGES = os.path.join(os.path.dirname(os.path.dirname(__file__)), "images")

@dataclass
class ImageInfo(object):
    versions: Dict[str, Dict[str, Any]]              # mapping of versions
    assign: Optional[Dict[str, Any]] = None       # optional assignments
    latest: Optional[str] = None                  # latest version -- use last 'versions' entry if not given
    dockerfile: Optional[str] = None

@dataclass
class Manifest(object):
    registry: str
    assign: Optional[Dict[str, Any]]
    images: Dict[str, ImageInfo]


def run(command, cwd=None, input=None):
    print(f"[bold]{cwd or '.'}$ {command}[/bold]")
    args = command.split()
    result = subprocess.run(args, cwd=cwd, input=input, text=True)
    if result.returncode:
        print(f"{command} failed with exit code {result.returncode}")
        sys.exit(1)
    return 0


@click.command()
@click.option('-m', '--manifest', type=click.Path(exists=True), 
                default=DEFAULT_MANIFEST, 
                help=f'Cargo manifest. Default is {DEFAULT_MANIFEST}.')
@click.option('-b', '--build', is_flag=True, help='Build only, do not push.')
@click.option('-p', '--push', is_flag=True, help='Push only, do not build.')
@click.option('-r', '--rebuild', is_flag=True, help='Ignore docker image caches (i.e. rebuild).')
@click.option('-a', '--all', is_flag=True, help='Build and/or push all images in manifest.')
@click.argument('imagenames', type=str, nargs=-1)
def build_cargo(manifest: str, build=False, push=False, all=False, rebuild=False, imagenames: List[str] = []):
    if not (build or push):
        build = push = True

    print(f"Loading manifest {manifest}")

    conf = OmegaConf.load(manifest)
    conf = OmegaConf.merge(OmegaConf.structured(Manifest), conf)

    print(f"Loaded {len(conf.images)} image entries")

    if all:
        imagenames = list(conf.images.keys())

    no_cache = "--no-cache" if rebuild else ""

    if not imagenames:
        print(f"Nothing to be done. Please specify some image names, or run with -a/-all.")
        return 0
    
    for image in imagenames:
        version = None
        if ':' in image:
            image, version = image.split(":", 1)
        if image not in conf.images:
            print(f"Unknown image '{image}'")
            return 1
        if version is not None and version not in conf.images[image].versions:
            print(f"Unknown image '{image}:{version}'")
            return 1
        
    global_vars = conf.assign or {}
    registry = global_vars.get('CULT_REGISTRY', 'quay.io/stimela2')
    CULT_VERSION = global_vars.get('CULT_VERSION')

    for image in imagenames:
        if ':' in image:
            image, version = image.split(":", 1)
            versions = [version]
            all_versions = False
        else:
            versions = conf.images[image].versions.keys()
            all_versions = True

        image_info = conf.images[image]
        image_vars = global_vars.copy()
        image_vars.update(IMAGE=image, **(image_info.assign or {}))
        image_vars.setdefault("CMD", image)

        path = os.path.join(BASE_IMAGES, image).format(**image_vars)
        latest = None

        for version in versions:
            version = version.format(**image_vars)
            version_info = image_info.versions[version]
            version_info.setdefault('VERSION', version)

            version_vars = image_vars.copy()
            version_vars.update(**version_info)
            
            dockerfile = version_info.get('dockerfile') or image_info.dockerfile or 'Dockerfile'
            dockerfile = dockerfile.format(**version_vars)
            image_version = version_info.VERSION.format(**version_vars)
            if image_version == "latest":
                latest = image_version = f"cc{CULT_VERSION}"
            else:
                image_version += f"-cc{CULT_VERSION}"

            dockerpath = os.path.join(path, dockerfile)
            print(f"[bold green]{image}:{image_version}[/bold green] defined by {dockerpath}")
            if not os.path.exists(dockerpath):
                print(f"  {dockerpath} doesn't exist")
                return 1
            # go build
            build_dir = os.path.dirname(dockerpath)
            full_image = f"{registry}/{image}:{image_version}"
            if build:
                content = open(dockerpath, "rt").read().format(**version_vars)
                run(f"docker build {no_cache} -t {full_image} -", cwd=build_dir, input=content)
            if push:
                run(f"docker push {full_image}", cwd=path)

        if all_versions and versions:
            if latest is None:
                latest = image_version  # use last version from loop above
                run(f"docker tag {registry}/{image}:{latest} {registry}/{image}:cc{CULT_VERSION}", cwd=build_dir)
                if push:
                    run(f"docker push {registry}/{image}:cc{CULT_VERSION}", cwd=path)
            run(f"docker tag {registry}/{image}:cc{CULT_VERSION} {registry}/{image}:latest", cwd=build_dir)
            if push:
                run(f"docker push {registry}/{image}:latest", cwd=path)



if __name__ == '__main__':
    build_cargo()


from scabha.cargo import Parameter
from typing import Dict, Any

def img_output(imagetype, desc, path, glob=True, must_exist=False):
    # reamp image type to output filename component
    if imagetype == "restored":
        imagetype = "image"
    implicit = f"{{current.prefix}}{path}-{imagetype}.fits"
    if glob:
        implicit = f"=GLOB({implicit})"
    return Parameter(
        info=f"{imagetype.capitalize()} {desc}",
        dtype="List[File]" if glob else "File",
        mkdir=True,
        implicit=implicit,
        must_exist=must_exist)   


def make_stimela_schema(params: Dict[str, Any], inputs: Dict[str, Parameter], outputs: Dict[str, Parameter]):
    """Augments a schema for stimela based on solver.terms"""

    # predict mode has no outputs
    if params.get('predict'):
        return inputs, outputs

    outputs = outputs.copy()

    # nchan -- if not an integer, assume runtime evaluation and >=2 then
    nchan  = params.get('nchan', 1)
    multichan = params.get('multi.chan', not isinstance(nchan, int) or nchan > 1)
    
    stokes = params.get('pol', "I").upper()
    # multi.stokes can be set explicitly
    multistokes = params.get('multi.stokes', stokes != "I")
    # if explicitly multi-pol, set Stokes list to "IQUV" if it's not explicitly set to something else
    if multistokes and stokes == "I":
        stokes = "IQUV"

    # ntime -- if not an integer, assume runtime evaluation and >=2 then
    ntime  = params.get('intervals-out', 1)
    multitime = params.get('multi.intervals', not isinstance(ntime, int) or ntime > 1)

    for imagetype in "dirty", "restored", "residual", "model":
        if imagetype == "dirty" or params.get('niter', 0) > 0:
            must_exist = True
        else:
            must_exist = False
        for st in stokes:
            # define name/description/filename components for this Stokes 
            if multistokes:
                st_name = f"{st.lower()}."
                st_name1 = f".{st.lower()}"
                st_desc = f"Stokes {st} "
                st_fname = f"-{st}"
            else:
                st_name = st_name1 = st_desc = st_fname = ""
            # now form up outputs
            if multitime:
                if multichan:
                    outputs[f"{imagetype}.{st_name}per-interval.per-band"] = img_output(imagetype,
                        f"{st_desc} images per time interval and band",
                        f"-t[0-9][0-9][0-9][0-9]-[0-9][0-9][0-9][0-9]{st_fname}", 
                        must_exist=must_exist)
                    outputs[f"{imagetype}.{st_name}per-interval.mfs"] = img_output(imagetype,
                        f"{st_desc} MFS image per time interval",
                        f"-t[0-9][0-9][0-9][0-9]-MFS{st_fname}",
                        must_exist=must_exist)
                else:
                    outputs[f"{imagetype}.{st_name}per-interval"] = img_output(imagetype,
                        f"{st_desc} image per time interval",
                        f"-t[0-9][0-9][0-9][0-9]{st_fname}",
                        must_exist=must_exist)

            else:
                if multichan:
                    outputs[f"{imagetype}.{st_name}per-band"] = img_output(imagetype,
                        f"{st_desc} images per band",
                        f"-[0-9][0-9][0-9][0-9]{st_fname}",
                        must_exist=must_exist)
                    outputs[f"{imagetype}.{st_name}mfs"] = img_output(imagetype,
                        f"{st_desc} MFS image",
                        f"-MFS{st_fname}", glob=False,
                        must_exist=must_exist)
                else:
                    outputs[f"{imagetype}{st_name1}"] = img_output(imagetype,
                        f"{st_desc} image",
                        f"{st_fname}", glob=False,
                        must_exist=must_exist)

    return inputs, outputs
